from starlette.datastructures import Headers
from contextvars import ContextVar
from starlette.types import ASGIApp, Receive, Scope, Send

HX_TARGET_CTX_KEY = "headers"

_hx_target_ctx_var: ContextVar[dict] = ContextVar(HX_TARGET_CTX_KEY, default=None)


def get_headers() -> dict[str:str]:
    return _hx_target_ctx_var.get()


class CustomRequestMiddleware:
    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ["http"]:
            await self.app(scope, receive, send)
            return
        headers = Headers(scope=scope)
        _headers = _hx_target_ctx_var.set(headers)

        await self.app(scope, receive, send)

        _hx_target_ctx_var.reset(_headers)



from typing import Literal
from uiwiz.element import Element


class Spinner(Element):
    _size: Literal["xs", "sm", "md", "lg"] = "sm"
    _type: Literal["spinner", "dots", "ring", "ball", "bars", "infinity"] = "ring"
    root_class: str = "loading "
    _classes: str = "loading-{0} loading-{1}"

    def __init__(self, spinner_for: Element = None) -> None:
        super().__init__("span")
        self.size = Spinner._size
        self.type = Spinner._type
        self.spinner_for = spinner_for
        self.__self__format()

    @property
    def spinner_for(self):
        return self._spinner_for

    @spinner_for.setter
    def spinner_for(self, element: Element) -> "Spinner":
        self._spinner_for = element
        if element:
            element.attributes["hx-indicator"] = f"#{self.id}"
            self.__self__format()

    def __self__format(self) -> None:
        clazz = Spinner._classes.format(self.type, self.size)
        if self.spinner_for:
            clazz += " htmx-indicator"
        self.classes(clazz)

    def extra_small(self) -> "Spinner":
        self.size = "xs"
        self.__self__format()
        return self

    def small(self) -> "Spinner":
        self.size = "sm"
        self.__self__format()
        return self

    def medium(self) -> "Spinner":
        self.size = "md"
        self.__self__format()
        return self

    def large(self) -> "Spinner":
        self.size = "lg"
        self.__self__format()
        return self

    def ring(self) -> "Spinner":
        self.type = "ring"
        self.__self__format()
        return self

    def spinner(self) -> "Spinner":
        self.type = "spinner"
        self.__self__format()
        return self

    def dots(self) -> "Spinner":
        self.type = "dots"
        self.__self__format()
        return self

    def ball(self) -> "Spinner":
        self.type = "ball"
        self.__self__format()
        return self

    def bars(self) -> "Spinner":
        self.type = "bars"
        self.__self__format()
        return self

    def infinity(self) -> "Spinner":
        self.type = "infinity"
        self.__self__format()
        return self



from typing import Callable, Optional
from uiwiz.element import Element


class Input(Element):
    root_class: str = "input "
    _classes: str = "input-bordered w-full max-w-xs"

    def __init__(
        self,
        placeholder: str = None,
        name=None,
        on_change: Optional[Callable] = None,
        target: Optional[Callable] = None,
        trigger: str = "input",
        swap: Optional[str] = "swap",
    ) -> None:
        """Input

        This element is used for input data

        :param placeholder: Text to show before input
        :param name: Used in the attributes name sent back in json
        :param on_change: function or method to call on change
        :param target: Function or id of the element to replace
        :param trigger: keyup, keydown, load and more see htmx docs
        """
        super().__init__("input")
        self.classes(Input._classes)
        self.attributes["name"] = name
        self.attributes["placeholder"] = placeholder
        self.attributes["autocomplete"] = "off"
        self.inline = True

        if on_change:
            self.event = {
                "func": on_change,
                "trigger": trigger,
                "target": target,
                "swap": swap,
            }



import inspect
from mimetypes import guess_type
import os
from pathlib import Path
from typing import Callable, Optional, Union

from fastapi import FastAPI, Response
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.requests import Request
from uiwiz.element import Element, Frame
from fastapi.middleware.gzip import GZipMiddleware
from uiwiz.header_middelware import CustomRequestMiddleware
import functools
import logging

logger = logging.getLogger("uiwiz")
logger.addHandler(logging.NullHandler())


class UiwizApp(FastAPI):
    app_paths = {}

    def __init__(
        self,
        toast_delay: int = 2500,
        error_classes: str = "alert bg-[#FF8080]",
        theme: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.toast_delay = toast_delay
        self.error_classes = error_classes
        if theme:
            self.theme = f"data-theme={theme}"
        else:
            self.theme = theme
        self.templates = Jinja2Templates(Path(__file__).parent / "templates")
        self.add_static_files("/static", Path(__file__).parent / "static")
        Frame.api = self
        self.add_middleware(CustomRequestMiddleware)
        self.add_middleware(GZipMiddleware)
        self.extensions: dict[str, Path] = {}

    def render(
        self,
        frame: Frame,
        request: Request,
        title: str,
        status_code: int = 200,
    ):
        html = frame.render()
        libs = frame.render_libs()
        ext = frame.render_ext()
        frame.del_stack()
        return self.templates.TemplateResponse(
            "default.html",
            {
                "request": request,
                "root_element": [html],
                "title": title,
                "theme": self.theme,
                "libs": libs,
                "ext": ext,
                "toast_delay": self.toast_delay,
                "error_classes": self.error_classes,
            },
            status_code,
            {"Cache-Control": "no-store", "X-uiwiz-Content": "page"},
        )

    def render_api(self, frame: Frame, status_code: int = 200):
        html = frame.render()
        frame.del_stack()
        return HTMLResponse(
            html,
            status_code,
            {"Cache-Control": "no-store", "X-uiwiz-Content": "page"},
        )

    def route_exists(self, path: str) -> None:
        return path in self.routes

    def remove_route(self, path: str) -> None:
        """Remove routes with the given path."""
        self.routes[:] = [r for r in self.routes if getattr(r, "path", None) != path]

    def add_static_files(self, url_path: str, local_directory: Union[str, Path]) -> None:
        self.mount(url_path, StaticFiles(directory=str(local_directory)))

    def register_extension(self, path: Path, prefix: str):
        _, filename = os.path.split(path)
        if filename in self.extensions:
            return
        self.extensions[filename] = path

        def get_extension(filename):
            if filename not in self.extensions:
                return Response(status_code=404)

            with open(self.extensions[filename]) as f:
                content = f.read()

            content_type, _ = guess_type(filename)
            return Response(content, media_type=content_type)

        self.get(prefix + "{filename}")(get_extension)

    def post(self, path: str, *args, **kwargs):
        s = super()

        def decorator(func: Callable, *args, **kwargs) -> Callable:
            self.app_paths[func] = path
            return s.post(path, *args, **kwargs)(func)

        return decorator

    def page(
        self,
        path: str,
        *args,
        title: Optional[str] = "uiwiz",
        favicon: Optional[str] = None,
    ) -> Callable:
        def decorator(func: Callable, *args, **kwargs) -> Callable:
            self.remove_route(path)
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            async def decorated(*dec_args, **dec_kwargs) -> Response:
                frame: Frame = Frame.get_stack()
                frame.app = self
                Element().classes("flex flex-col h-screen")
                request = dec_kwargs["request"]
                # NOTE cleaning up the keyword args so the signature is consistent with "func" again
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result
                if isinstance(result, Response):  # NOTE if setup returns a response, we don't need to render the page
                    return result

                return self.render(frame, request, title)

            params = [p for p in inspect.signature(func).parameters.values()]
            if "request" not in {p.name for p in params}:
                request = inspect.Parameter(
                    "request",
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=Request,
                )
                params.insert(0, request)
            decorated.__signature__ = inspect.Signature(params)

            self.app_paths[decorated] = path

            return self.get(path)(decorated)

        return decorator

    def ui(self, path: str) -> Callable:
        def decorator(func: Callable) -> Callable:
            self.remove_route(path)
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs) -> Response:
                frame = Frame.get_stack()
                # NOTE cleaning up the keyword args so the signature is consistent with "func" again
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result
                if isinstance(result, Response):  # NOTE if setup returns a response, we don't need to render the page
                    return result

                return self.render_api(frame)

            request = inspect.Parameter("request", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=Request)
            params = [p for p in inspect.signature(func).parameters.values()]
            for p in params:
                if p.annotation == inspect.Signature.empty:
                    p._annotation = Request
            if "request" not in {p.name for p in params}:
                params.insert(0, request)
            decorated.__signature__ = inspect.Signature(params)

            if not self.route_exists(path):
                self.app_paths[decorated] = path
            return self.post(path)(decorated)

        return decorator



from typing import Callable, Optional, Union
from uiwiz.element import Element


class Button(Element):
    root_class: str = "btn "

    def __init__(self, title: str) -> None:
        super().__init__(tag="button")
        self.content = title
        self.classes()
        self.inline = True

    def on_click(
        self,
        func: Optional[Callable] = None,
        target: Union[Callable, str, Element] = None,
        swap: Optional[str] = None,
    ) -> "Button":
        self.event = {
            "func": func,
            "trigger": "click",
            "target": target,
            "swap": swap,
        }
        return self



from typing import Optional
from uiwiz.element import Element
from uiwiz.elements.extensions.bindable import Bindable


class Label(Bindable):
    root_class: str = "label "

    def __init__(self, text: Optional[str] = None, for_: Element = None) -> None:
        super().__init__(tag="label")
        self.inline = True
        if text:
            self.content = text
        if for_:
            self.attributes["for"] = for_.id
        self.classes()

    def set_text(self, text: str) -> "Label":
        self.content = text
        return self



from typing import Callable, Optional
from uiwiz.element import Element


class TextArea(Element):
    root_class: str = "textarea "
    _classes: str = "textarea-bordered"

    def __init__(
        self,
        placeholder: str = None,
        name="input",
        on_change: Callable = None,
        target: Callable = None,
        trigger: str = "input",
        swap: str = "swap",
    ) -> None:
        super().__init__("textarea")
        self.attributes["name"] = name
        self.attributes["placeholder"] = placeholder
        self.inline = True
        self.classes(TextArea._classes)

        if on_change:
            self.event = {
                "func": on_change,
                "trigger": trigger,
                "target": target,
                "swap": swap,
            }

    @property
    def placeholder(self) -> Optional[str]:
        return self.attributes["placeholder"]



from uiwiz.element import Element


class Link(Element):
    root_class: str = "link "
    _classes: str = "link-hover"

    def __init__(self, text: str, link: str) -> None:
        super().__init__("a")
        self.content = text
        self.attributes["href"] = link
        self.classes(Link._classes)



import asyncio
import os
from pathlib import Path
from typing import Callable, Optional, Union
import random
from uiwiz.header_middelware import get_headers
from uiwiz.event import Event

# https://developer.mozilla.org/en-US/docs/Glossary/Void_element
VOID_ELEMENTS = [
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
]

MAX_RAND_ID = 999999999


class Frame:
    stacks: dict[int, "Frame"] = {}
    api = None

    def __init__(self) -> None:
        self.root_element: Optional[Element] = None
        self.current_element: Optional[Element] = None
        self.oob_elements: list[Element] = []
        self.used_hx_headers: bool = False
        self.id_count: int = 0  # used for element id
        self.scripts: list[str] = []
        self.libraries: list[str] = []
        self.extensions: set[str] = set()
        # TODO: Implement
        self.last_id = None

    def get_id(self) -> str:
        headers = get_headers()
        if swap := headers.get("hx-swap"):
            target_id = self.last_id if self.last_id else headers.get("hx-target")

            if self.used_hx_headers is False:
                if swap in ["outerHTML", "this"]:
                    self.last_id = target_id
                else:
                    generator = random.Random(target_id)
                    self.last_id = f"a-{generator.randint(10000, MAX_RAND_ID)}"
                self.used_hx_headers = True
                return self.last_id

            generator = random.Random(target_id)
            self.last_id = f"a-{generator.randint(10000, MAX_RAND_ID)}"
            return self.last_id

        return f"a-{self.id_count}"

    def remove_frame(self, element: "Element") -> None:
        if element is None:
            if self.root_element:
                self.root_element.stack = None
                self.remove_frame(self.root_element)
        else:
            element.stack = None
            for child in element.children:
                self.remove_frame(child)

    def render(self) -> str:
        if self.root_element:
            return self.root_element.render()
        if self.oob_elements:
            output = ""
            el: Element
            for el in self.oob_elements:
                output += el.render_top_level()
            return output
        return ""

    def render_libs(self) -> str:
        return self.root_element.render_js(self.libraries)

    def render_ext(self) -> str:
        return self.root_element.render_js(self.extensions)

    def add_extension(self, path: Path):
        _, filename = os.path.split(path)
        prefix = "/_static/ext/"
        endpoint = prefix + filename
        self.extensions.add(endpoint)
        Frame.api.register_extension(path, prefix)

    @classmethod
    def get_stack(cls) -> "Frame":
        _id = get_task_id()
        if _id not in cls.stacks:
            cls.stacks[_id] = Frame()
        return cls.stacks[_id]

    @classmethod
    def del_stack(cls):
        cls.get_stack().remove_frame(None)
        del cls.stacks[get_task_id()]


def get_task_id() -> int:
    try:
        return id(asyncio.current_task())
    except RuntimeError:
        return 0


class Element:
    def __init__(
        self,
        tag="div",
        indent_level=2,
        content="",
        render_html=True,
        libraries: list[str] = [],
        extension: Path = None,
        oob: bool = False,
    ) -> None:
        self.stack = Frame.get_stack()
        self.stack.libraries.extend(libraries)
        if extension:
            self.stack.add_extension(extension)
        self.attributes: dict[str, str] = {}
        self.attributes["id"] = self.stack.get_id()
        self.stack.id_count += 1
        self.tag: str = tag

        self.event: Event = {}
        self.parent_element: Element = None
        self.children: list[Element] = []
        self.script: Optional[str] = None
        self.render_html: bool = render_html
        self.target: Optional[str] = None
        self.inline: bool = False

        self.content: str = content
        self.indent: int = indent_level
        self.oob: bool = oob

        if self.oob:
            self.attributes["hx-swap-oob"] = "true"
            self.stack.oob_elements.append(self)

        if self.stack.root_element is None:
            self.stack.root_element = self
            self.stack.current_element = self
        else:
            self.stack.current_element.children.append(self)
            self.parent_element = self.stack.current_element

    def __enter__(self):
        self.stack.current_element = self
        return self

    def __exit__(self, *_):
        self.stack.current_element = self.parent_element

    @property
    def id(self):
        return self.attributes["id"]

    @property
    def name(self):
        return self.attributes["name"]

    @property
    def value(self):
        return self.attributes["value"]

    @value.setter
    def value(self, value):
        self.attributes["value"] = value

    @property
    def is_void_element(self) -> bool:
        if self.tag in VOID_ELEMENTS:
            return True
        return False

    def get_classes(self) -> str:
        return self.attributes["class"]

    def classes(self, input: str = ""):
        self.attributes["class"] = getattr(self.__class__, "root_class", "") + input
        return self

    def render(self, render_script: bool = True) -> str:
        output = self.render_top_level(render_script)
        output += self.render_oob()
        return output

    def render_top_level(self, render_script: bool = True, render_oob: bool = False) -> str:
        lst = []
        lst.append(self.render_self(render_oob=render_oob))
        if render_script:
            for script in self.stack.scripts:
                lst.append(
                    """
                    <script>
                    (function() {
                    %s
                    }());
                    </script>
                    """
                    % script
                )
        return "".join(lst)

    def render_self(self, render_oob: bool = False) -> str:
        if self.oob and render_oob is False:
            return ""

        self.before_render()
        lst = []
        if self.render_html:
            self.add_event_to_attributes()

            lst.append("<%s %s>" % (self.tag, self.dict_to_attrs()))
            lst.append(str(self.content))
            lst.extend([child.render_self() if child.oob is False else "" for child in self.children])

            if not self.is_void_element:
                lst.append("</%s>" % self.tag)

        return "".join(lst)

    def render_oob(self) -> str:
        lst = []
        el: Element
        for el in self.stack.oob_elements:
            lst.append(el.render_top_level(render_oob=True))
        return "".join(lst)

    def before_render(self):
        pass

    def add_event_to_attributes(self):
        if self.event == {}:
            return

        self.attributes["hx-target"] = self.get_target(self.event.get("target"))
        self.attributes["hx-swap"] = self.event.get("swap") if self.event.get("swap") is not None else "outerHTML"

        func = self.event["func"]
        endpoint = Frame.api.app_paths.get(func)
        if endpoint is None:
            endpoint = f"/_uiwiz/hash/{func.__hash__()}"
            Frame.api.ui(endpoint)(func)

        self.attributes["hx-post"] = endpoint
        if self.event.get("trigger"):
            self.attributes["hx-trigger"] = self.event["trigger"]

        if vals := self.event.get("vals"):
            self.attributes["hx-vals"] = vals
        if include := self.event.get("include"):
            self.attributes["hx-include"] = include
        if hx_encoding := self.event.get("hx-encoding"):
            self.attributes["hx-encoding"] = hx_encoding
        else:
            self.attributes["hx-ext"] = "json-enc"

    def get_target(self, target: Union[Callable, str, "Element", None]) -> str:
        _target = "this"
        if target is None:
            return _target

        if isinstance(target, Callable):
            return "#%s" % str(target())

        if isinstance(target, Element):
            return "#%s" % target.id

        if target != "this" and "#" not in target:
            return "#" + target
        else:
            return target

    def render_js(self, lst_js: list[str]) -> str:
        lst = set()
        for lib in lst_js:
            lst.add('<script src="%s"></script>' % lib)
        return "".join(lst)

    def __str__(self) -> str:
        return self.render()

    def set_frame(self, frame: Frame):
        self.stack = frame
        for child in self.children:
            child.set_frame(frame)

    def set_frame_and_root(self) -> None:
        self.set_frame(Frame.get_stack())
        self.stack.root_element = self

    def dict_to_attrs(self):
        ATTR_NO_VALUE = object()
        return " ".join(
            (key if value is ATTR_NO_VALUE else '%s="%s"' % (key, value)) for key, value in self.attributes.items()
        )



from typing import Literal
from anyio import Path

_type = Literal["info", "error", "success", "warning", "menu", None]


def get_svg(svg: Literal["info", "error", "success", "warning", "menu"]) -> str:
    path = Path(__file__).parent / (svg + ".svg")
    with open(path, "r") as f:
        return f.read()



from typing import Callable, Optional, Union
from uiwiz.element import Element


class Form(Element):
    root_class: str = "col "

    def __init__(self) -> None:
        super().__init__("form")
        self.classes()

    def on_submit(
        self,
        func: Optional[Callable] = None,
        target: Union[Callable, str, Element] = None,
        swap: Optional[str] = "beforeend",
    ):
        self.event = {
            "func": func,
            "trigger": "submit",
            "target": target,
            "swap": swap,
        }
        return self
